/*
 * Copyright 2012, 2017 INFN
 *
 * Licensed under the EUPL, Version 1.2 or – as soon they
 * will be approved by the European Commission - subsequent
 * versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the
 * Licence.
 * You may obtain a copy of the Licence at:
 *
 * https://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in
 * writing, software distributed under the Licence is
 * distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied.
 * See the Licence for the specific language governing
 * permissions and limitations under the Licence.
 */

#pragma once

namespace bson {
namespace mutablebson {

    inline Element Element::operator[](size_t n) const {
        return findNthChild(n);
    }

    inline Element Element::operator[](const StringData& name) const {
        return findFirstChildNamed(name);
    }

    inline double Element::getValueDouble() const {
        dassert(hasValue() && isType(bson::NumberDouble));
        return getValue()._numberDouble();
    }

    inline StringData Element::getValueString() const {
        dassert(hasValue() && isType(bson::String));
        return getValueStringOrSymbol();
    }

    inline BSONObj Element::getValueObject() const {
        dassert(hasValue() && isType(bson::Object));
        return getValue().Obj();
    }

    inline BSONArray Element::getValueArray() const {
        dassert(hasValue() && isType(bson::Array));
        return BSONArray(getValue().Obj());
    }

    inline bool Element::isValueUndefined() const {
        return isType(bson::Undefined);
    }

    inline OID Element::getValueOID() const {
        dassert(hasValue() && isType(bson::jstOID));
        return getValue().__oid();
    }

    inline bool Element::getValueBool() const {
        dassert(hasValue() && isType(bson::Bool));
        return getValue().boolean();
    }

    inline Date_t Element::getValueDate() const {
        dassert(hasValue() && isType(bson::Date));
        return getValue().date();
    }

    inline bool Element::isValueNull() const {
        return isType(bson::jstNULL);
    }

    inline StringData Element::getValueSymbol() const {
        dassert(hasValue() && isType(bson::Symbol));
        return getValueStringOrSymbol();
    }

    inline int32_t Element::getValueInt() const {
        dassert(hasValue() && isType(bson::NumberInt));
        return getValue()._numberInt();
    }

    inline OpTime Element::getValueTimestamp() const {
        dassert(hasValue() && isType(bson::Timestamp));
        return getValue()._opTime();
    }

    inline int64_t Element::getValueLong() const {
        dassert(hasValue() && isType(bson::NumberLong));
        return getValue()._numberLong();
    }

    inline bool Element::isValueMinKey() const {
        return isType(bson::MinKey);
    }

    inline bool Element::isValueMaxKey() const {
        return isType(bson::MaxKey);
    }

    inline bool Element::ok() const {
        dassert(_doc != NULL);
        return _repIdx <= kMaxRepIdx;
    }

    inline Document& Element::getDocument() {
        return *_doc;
    }

    inline const Document& Element::getDocument() const {
        return *_doc;
    }

    inline bool Element::isType(BSONType type) const {
        return (getType() == type);
    }

    inline Element::RepIdx Element::getIdx() const {
        return _repIdx;
    }

    inline Element::Element(Document* doc, RepIdx repIdx)
        : _doc(doc)
        , _repIdx(repIdx) {
        dassert(_doc != NULL);
    }

    inline StringData Element::getValueStringOrSymbol() const {
        const BSONElement value = getValue();
        const char* str = value.valuestr();
        const size_t size = value.valuestrsize() - 1;
        return StringData(str, size);
    }

    inline bool operator==(const Element& l, const Element& r) {
        return (l._doc == r._doc) && (l._repIdx == r._repIdx);
    }

    inline bool operator!=(const Element& l, const Element& r) {
        return !(l == r);
    }


} // namespace mutablebson
} // namespace bson
