/*
 * Copyright 2012, 2017 INFN
 *
 * Licensed under the EUPL, Version 1.2 or – as soon they
 * will be approved by the European Commission - subsequent
 * versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the
 * Licence.
 * You may obtain a copy of the Licence at:
 *
 * https://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in
 * writing, software distributed under the Licence is
 * distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied.
 * See the Licence for the specific language governing
 * permissions and limitations under the Licence.
 */

#pragma once

#include <iosfwd>
#include <string>

#ifndef _WIN32
#include <unistd.h>
#endif

#include <chaos/common/bson/platform/cstdint.h>
#include <chaos/common/bson/platform/hash_namespace.h>

namespace bson {

#ifdef _WIN32
    typedef DWORD NativeProcessId;
#else
    typedef pid_t NativeProcessId;
#endif

    /**
     * Platform-independent representation of a process identifier.
     */
    class ProcessId {
    public:
        /**
         * Gets the process id for the currently executing process.
         */
        static ProcessId getCurrent();

        /**
         * Constructs a ProcessId from a NativeProcessId.
         */
        static inline ProcessId fromNative(NativeProcessId npid) { return ProcessId(npid); }

        /**
         * Constructs a ProcessId with an unspecified value.
         */
        ProcessId() {}

        /**
         * Gets the native process id corresponding to this ProcessId.
         */
        NativeProcessId toNative() const { return _npid; }

        /**
         * Represents this process id as a signed 64-bit integer.
         *
         * This representation will faithfully serialize and format to text files, but is at least
         * twice the number of bits needed to uniquely represent valid process numbers on supported
         * OSes..
         */
        int64_t asInt64() const;

        /**
         * Similar to asInt64(), for compatibility with code that uses "long long" to mean 64-bit
         * signed integer.
         */
        long long asLongLong() const;

        /**
         * Represents this process id as an unsigned 32-bit integer.
         *
         * This representation will contain all of the bits of the native process id, but may not
         * serialize or format to text files meaningfully.
         */
        uint32_t asUInt32() const { return static_cast<uint32_t>(_npid); }

        /**
         * Provides a string representation of the pid.
         */
        std::string toString() const;

        bool operator==(const ProcessId other) const { return _npid == other._npid; }
        bool operator!=(const ProcessId other) const { return _npid != other._npid; }
        bool operator<(const ProcessId other) const { return _npid < other._npid; }
        bool operator<=(const ProcessId other) const { return _npid <= other._npid; }
        bool operator>(const ProcessId other) const { return _npid > other._npid; }
        bool operator>=(const ProcessId other) const { return _npid >= other._npid; }

    private:
        explicit ProcessId(NativeProcessId npid): _npid(npid) {}

        NativeProcessId _npid;
    };

    std::ostream& operator<<(std::ostream& os, ProcessId pid);

}  // namespace bson

BSON_MONGO_HASH_NAMESPACE_START
template<> struct hash< ::bson::ProcessId > {
    size_t operator()(const ::bson::ProcessId pid) const {
        return hash< ::bson::uint32_t >()(pid.asUInt32());
    }
};
BSON_MONGO_HASH_NAMESPACE_END
